//*******************************************************
//	NS73M FM transmitter controler
//
//	MCU:	ATtiny13V
//			hfuse:FDh	BOD 1.8V
//			lfuse:21h	0.6MHz
//
//	FM transmitter module : NS73M (NIIGATA SEIMITSU CO., LTD.)
//
//	Compiler:	AVR-GCC		4.2.2
//				AVR-LIBC	1.4.7
//				binutils	2.18
//
//	2008.8.9	N̎ԑ҂𒷂
//	2008.8.7	gɉ CEXlω悤ɂ
//	2008.8.5	gvZÔPĂƐmɂ
//	2008.8.4	vGt@VX 50usɏCANԑ҂500ms
//	2008.8.3	KAWAKAMI Yukio
//*******************************************************

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<avr/eeprom.h>

#define		NOP		asm volatile("nop\n"::)
#define		WDR		asm volatile("wdr\n"::)
#define		SLEEP	asm volatile("sleep\n"::)

typedef	unsigned char	uchar;
typedef	unsigned int	uint;
typedef	unsigned long	ulong;

#define		EEPROM_FREQ	(uint16_t *)1


// |[gݒ(PB)
#define		SDA		4	// PB
#define		SCL		3	// PB
#define		NC1		2
#define		LED		1
#define		NC2		0

// TWI(I2C)̃}Nݒ
#define		SCL_H	\
	DDRB &= ~(1<<SCL);	\
	PORTB |= (1<<SCL)

#define		SCL_L	\
	PORTB &= ~(1<<SCL);		\
	DDRB |= (1<<SCL)

#define		SDA_H	\
	DDRB &= ~(1<<SDA);	\
	PORTB |= (1<<SDA)

#define		SDA_L	\
	PORTB &= ~(1<<SDA);		\
	DDRB |= (1<<SDA)

// SCL=HɂȂ܂ő҂
#define		WAIT_SCL_H	while(!(PINB & (1<<SCL)))

void TWI_start(void);
void TWI_send(uchar data);
uchar TWI_recv(uchar ack);
void TWI_stop(void);

void set_data(uchar reg, uchar data);
void wait_100ms(uchar n);

//*******************************************************
//	C[`
//*******************************************************
int main (void) {
	ulong	freq;
	uchar	cex;

	cli();	// 荞݋֎~

	// |[gݒ
	PORTB = (1<<NC1)|(1<<NC2)|(1<<SCL)|(1<<SDA);	// vAbv
	DDRB = (1<<LED)|(1<<SDA)|(1<<SCL);	// o̓|[g

	// X[vݒ
	MCUCR = (1<<SE)|(1<<SM1);	// X[v, p[_E[h

	// ^C}0ݒ
	TCCR0A = 0;
	TCCR0B = (1<<CS02)|(1<<CS00);	// 0.6MHz/1024=586Hz

	// 1bԑ҂
	wait_100ms(10);

	// EEPROMgǂݍ
	freq = (ulong)eeprom_read_word(EEPROM_FREQ);
	if ((freq < 7600)||(freq > 10800)){
		freq = 8820;	// ftHg 88.2MHz
	}

	// CEXlI
	if (freq < 9000){
		cex = 3;
	} else if (freq < 9600){
		cex = 2;
	} else if (freq < 10200){
		cex = 1;
	} else {
		cex = 0;
	}

	// WX^ݒlvZ
	freq *= 10000;
	freq += 304000;
	freq += 8192/2;
	freq /= 8192;
	
	// NS73M I2CZbg
	TWI_start();
	TWI_send(0xFF);
	TWI_start();
	TWI_stop();

	set_data(0x0E, 0x05);	// Zbg
	set_data(0x01, 0xB4);	// Pilot ON, Forced subcarrier
	set_data(0x02, 0x06);	// ULD ON, 1mW
	set_data(0x03, (uchar)freq);		// g
	set_data(0x04, (uchar)(freq >> 8));	// g
	set_data(0x08, 0x18 + cex);	// Local OSC Band
	set_data(0x00, 0x05);	// input 100mV, Pre-emphasis 50us, Power ON, Mute

	wait_100ms(1);

	set_data(0x0E, 0x05);	// Zbg
	set_data(0x06, 0x1E);	// Charge pump

	wait_100ms(2);

	set_data(0x00, 0x01);	// Audio input 100mV, Pre-emphasis 50us, Power ON

	PORTB |= (1<<LED);		// LED_

	SLEEP;	// ͐Q

	for (;;){
	}
}


//*******************************************************
//	100msPʂŎԑ҂
//*******************************************************
void wait_100ms(uchar n){

	while(n > 0){
		// 0.1bԑ҂
		TCNT0 = 256 - 59;
		TIFR0 |= (1<<TOV0);	// I[o[t[tONA
		while(!(TIFR0 & (1<<TOV0)));
		--n;
	}
}


//************************************************************
//	WX^Ƀf[^Zbg
//************************************************************
void set_data(uchar reg, uchar data){

	TWI_start();
	TWI_send(0xCC);		// AhX + Write
	TWI_send(reg);		// WX^
	TWI_send(data);		// f[^
	TWI_stop();
}


//************************************************************
//	TWI Jn
//************************************************************
void TWI_start(void){

	SDA_H;
	SCL_H;
	SDA_L;
}


//************************************************************
//	TWI 8bit M
//************************************************************
void TWI_send(uchar data){
	char	bit;

	SCL_L;
	for (bit = 0; bit < 8; bit++){
		if (data & 0x80){
			SDA_H;
		} else {
			SDA_L;
		}
		SCL_H;
		WAIT_SCL_H;		// SCL=HɂȂ܂ő҂
		data <<= 1;
		SCL_L;
	}

	SDA_H;
	SCL_L;

	// ACKM e͖
	SCL_H;
	NOP;
	SCL_L;
}


//************************************************************
//	TWI 8bit M
//************************************************************
uchar TWI_recv(uchar nack){
	char	bit;
	uchar	data;

	SDA_H;
	SCL_L;

	data = 0;
	for (bit = 0; bit < 8; bit++){
		SCL_H;
		WAIT_SCL_H;		// SCL=HɂȂ܂ő҂
		data <<= 1;
		if (PINB & (1<<SDA)){
			data |= 1;
		}
		SCL_L;
	}

	// ACK/NACK M
	if (nack){
		SCL_L;
		SCL_H;
		SDA_H;
	} else {
		SDA_L;
		SCL_L;
		SCL_H;
	}

	SCL_L;
	SDA_H;

	return data;
}


//************************************************************
//	TWI ~
//************************************************************
void TWI_stop(void){

	SDA_L;
	NOP;
	SCL_H;
	NOP;
	SDA_H;
}


